#include <stdio.h>
#include <stdlib.h>
#include <portaudio.h>

#include "ugens/sinosc.h"
#include "ugens/mult.h"

struct graph_node_s {
    void *ugen_instance;
    void (*tick)(void *);
    void (*destroy)(void *);
    struct graph_node_s *next;
};

typedef struct graph_node_s graph_node_t;

typedef struct
{
    graph_node_t *graphHead;
    double out1;
    double out2;
} paData;

static int paCallback( const void *inputBuffer, void *outputBuffer,
                           unsigned long framesPerBuffer,
                           const PaStreamCallbackTimeInfo* timeInfo,
                           PaStreamCallbackFlags statusFlags,
                           void *userData )
{
    /* Cast data passed through stream to our structure. */
    paData *data = (paData*)userData;
    float *out = (float*)outputBuffer;
    float *in = (float*)inputBuffer;
    unsigned int i;
    (void) inputBuffer; /* Prevent unused variable warning. */
    for( i=0; i<framesPerBuffer; i++ ) {
        graph_node_t *node = data->graphHead;
        while (node) {
            node->tick(node->ugen_instance);
            node = node->next;
        }
        *out++ = data->out1;
        *out++ = data->out2;
    }
    return 0;
}



int main(void)
{
    PaError err;
    int sr = 44100;
    paData data;
    data.out1 = data.out2 = 0;

    graph_node_t *node1 = (graph_node_t *) malloc(sizeof(graph_node_t));
    graph_node_t *node2 = (graph_node_t *) malloc(sizeof(graph_node_t));
    graph_node_t *node3 = (graph_node_t *) malloc(sizeof(graph_node_t));

    node1->ugen_instance = sinosc_construct(sr, 440, 0.3);
    node1->tick = sinosc_process;
    node1->destroy = sinosc_destroy;

    node2->ugen_instance = sinosc_construct(sr, 660, 0.3);
    node2->tick = sinosc_process;
    node2->destroy = sinosc_destroy;

    node3->ugen_instance = mult_construct();
    node3->tick = mult_process;
    node3->destroy = mult_destroy;
    node3->next = NULL;
    mult_set_output(node3->ugen_instance, 0, &data.out1);

//    sinosc_set_output(node1->ugen_instance, 0, mult_get_input(node3, 0));
//    sinosc_set_output(node2->ugen_instance, 0, mult_get_input(node3, 1));

    sinosc_set_output(node1->ugen_instance, 0, &data.out1);
    sinosc_set_output(node2->ugen_instance, 0, &data.out2);

    node1->next = node2;
    node2->next = NULL;

    data.graphHead = node1;

    err = Pa_Initialize();
    if( err != paNoError )
        printf( "PortAudio error: %s\n", Pa_GetErrorText( err ) );

    PaStream *stream;
    err = Pa_OpenDefaultStream( &stream,
                                2, /* no input channels */
                                2, /* stereo output */
                                paFloat32, /* 32 bit floating point output */
                                sr,
                                256, /* frames per buffer */
                                paCallback, /* this is your callback function */
                                &data ); /*This is a pointer that will be passed to
   your callback*/
    if( err != paNoError )
        printf( "PortAudio error: %s\n", Pa_GetErrorText( err ) );

    err = Pa_StartStream( stream );
    if( err != paNoError )
        printf( "PortAudio error: %s\n", Pa_GetErrorText( err ) );

    Pa_Sleep(5*1000);

    err = Pa_StopStream( stream );
    if( err != paNoError )
        printf( "PortAudio error: %s\n", Pa_GetErrorText( err ) );

    err= Pa_CloseStream(stream);
    if( err != paNoError )
        printf( "PortAudio error: %s\n", Pa_GetErrorText( err ) );

    err = Pa_Terminate();
    if( err != paNoError )
        printf( "PortAudio error: %s\n", Pa_GetErrorText( err ) );

    while (data.graphHead) {
        data.graphHead->destroy(data.graphHead->ugen_instance);
        graph_node_t *oldNode = data.graphHead;
        data.graphHead = data.graphHead->next;
        free(oldNode);
    }

    printf("Hello World!\n");
    return 0;
}

