
#include <stdlib.h>
#include <math.h>

#include "sinosc.h"

#define PHSMASK64 0x0FFFFFFFFFFFFFFFL
#define MAXLEN64  0x1000000000000000L

void *sinosc_construct(int sr, double freq, double amp)
{
    int i;
    SINOSC_DATA *data = (SINOSC_DATA *) malloc(sizeof(SINOSC_DATA));
    data->amp = (double *) malloc(sizeof(double));
    *data->amp = amp;
    data->freq = (double *) malloc(sizeof(double));
    *data->freq = freq;
    data->sr = sr;
    data->phs = 0;
    data->lobits = log2(MAXLEN64/SINOSC_BUFFER_SIZE);
    sinosc_set_param(data, SINOSC_FREQ, freq);
    sinosc_set_param(data, SINOSC_AMP, amp);

    for (i = 0; i < SINOSC_BUFFER_SIZE; i++) {
        data->buffer[i] = sin(2.0 * M_PI * i / (double) SINOSC_BUFFER_SIZE);
    }
    return (void *) data;
}


void sinosc_process(void *data)
{
    SINOSC_DATA *sinosc = (SINOSC_DATA *) data;
    double out;
    sinosc_set_param(sinosc, SINOSC_FREQ, *sinosc->freq);
    out = sinosc->buffer[sinosc->phs >> sinosc->lobits];
    sinosc->phs = (sinosc->phs + sinosc->inc) & PHSMASK64;
    *sinosc->out = out * *sinosc->amp;
}

void sinosc_destroy(void *data)
{
    SINOSC_DATA *sinosc = (SINOSC_DATA *) data;
    free(sinosc->amp);
    free(sinosc->freq);
    free(sinosc);
}

void sinosc_set_param(void *data, int index, double value)
{
    SINOSC_DATA *sinosc = (SINOSC_DATA *) data;
    if (index >= SINOSC_NUM_PARAMS || index < 0) {
        return;
    }

    switch(index) {
    case SINOSC_FREQ:
        sinosc->inc = (u_int64_t) lrint(value * (double) MAXLEN64/ sinosc->sr);
        break;
    case SINOSC_AMP:
        *sinosc->amp = value;
        break;
    }
}


double *sinosc_get_input(void *data, int index)
{
    SINOSC_DATA *sinosc = (SINOSC_DATA *) data;
    switch (index) {
    case SINOSC_FREQ:
        return sinosc->freq;
    case SINOSC_AMP:
        return sinosc->amp;
    }
    return NULL;
}


void sinosc_set_output(void *data, int index, double *out_buffer)
{
    SINOSC_DATA *sinosc = (SINOSC_DATA *) data;
    switch (index) {
    case 0:
        sinosc->out = out_buffer;
        break;
    }
}
